"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const sequelize_1 = require("sequelize");
const versioning_1 = require("../utils/versioning");
const string_1 = require("../utils/string");
const models_1 = require("../services/models");
const parentPrototype = versioning_1.majorVersion === 3 ? sequelize_1.Instance.prototype : sequelize_1.Model.prototype;
class BaseModel {
    static extend(target) {
        // PROTOTYPE MEMBERS
        // --------------------------
        // copies all prototype members of this to target.prototype
        Object
            .getOwnPropertyNames(this.prototype)
            .forEach(name => target.prototype[name] = this.prototype[name]);
        // STATIC MEMBERS
        // --------------------------
        // copies all static members of this to target
        Object
            .keys(this)
            .forEach(name => target[name] = this[name]);
        // Creates proxies for pre conforming include options; see "preConformIncludes"
        Object
            .keys(this.toPreConformIncludeMap)
            .forEach(key => {
            const superFn = target[key];
            target[key] = function (...args) {
                const options = args[BaseModel.toPreConformIncludeMap[key]];
                if (options) {
                    args[BaseModel.toPreConformIncludeMap[key]] = models_1.preConformIncludes(options, this);
                }
                return superFn.call(this, ...args);
            };
        });
    }
    static prepareInstantiationOptions(options, source) {
        options = models_1.preConformIncludes(options, source);
        if (!('isNewRecord' in options))
            options.isNewRecord = true;
        // TODO@robin has to be validated: necessary?
        // options = _.extend({
        //   isNewRecord: true,
        //   $schema: this.$schema,
        //   $schemaDelimiter: this.$schemaDelimiter
        // }, options || {});
        const staticMethodPrefix = versioning_1.majorVersion === 3 ? '$' : '_';
        // preventing TypeError: Cannot read property 'indexOf' of undefined(=includeNames)
        if (!options['includeNames'])
            options['includeNames'] = [];
        if (!options['includeValidated']) {
            sequelize_1.Model[staticMethodPrefix + 'conformOptions'](options, source);
            if (options.include) {
                sequelize_1.Model[staticMethodPrefix + 'expandIncludeAll'].call(source, options);
                sequelize_1.Model[staticMethodPrefix + 'validateIncludedElements'].call(source, options);
            }
        }
        return options;
    }
    /**
     * Adds relation between specified instances and source instance
     */
    $add(propertyKey, instances, options) {
        // TODO@robin find a way to add values to the target(this) datavalues
        // const dataValues = this['dataValues']; // this will not work correctly
        //
        // if (!dataValues[propertyKey]) dataValues[propertyKey] = [];
        //
        // dataValues[propertyKey].push(value);
        return this['add' + string_1.capitalize(propertyKey)](instances, options);
    }
    ;
    /**
     * Sets relation between specified instances and source instance
     * (replaces old relations)
     */
    $set(propertyKey, instances, options) {
        // TODO@robin find a way to add values to the target(this) datavalues
        // this['dataValues'][propertyKey] = args[0]; // this will not work correctly
        return this['set' + string_1.capitalize(propertyKey)](instances, options);
    }
    ;
    /**
     * Returns related instance (specified by propertyKey) of source instance
     */
    $get(propertyKey, options) {
        return this['get' + string_1.capitalize(propertyKey)](options);
    }
    ;
    /**
     * Counts related instances (specified by propertyKey) of source instance
     */
    $count(propertyKey, options) {
        return this['count' + string_1.capitalize(propertyKey)](options);
    }
    ;
    /**
     * Creates instances and relate them to source instance
     */
    $create(propertyKey, values, options) {
        return this['create' + string_1.capitalize(propertyKey)](values, options);
    }
    ;
    /**
     * Checks if specified instances is related to source instance
     */
    $has(propertyKey, instances, options) {
        return this['has' + string_1.capitalize(propertyKey)](instances, options);
    }
    ;
    /**
     * Removes specified instances from source instance
     */
    $remove(propertyKey, instances, options) {
        return this['remove' + string_1.capitalize(propertyKey)](instances, options);
    }
    ;
    /**
     * Pre conforms includes
     *
     * SEE DETAILS FOR ACTUAL FUNCTIONALITY ON DECLARATION FILE
     */
    reload(options) {
        return parentPrototype.reload.call(this, models_1.preConformIncludes(options, this));
    }
    ;
}
/**
 * Indicates which static methods of Model has to be proxied,
 * to prepare include option to automatically resolve alias;
 * The index represents the index of the options of the
 * corresponding method parameter
 */
BaseModel.toPreConformIncludeMap = {
    bulkBuild: 1,
    build: 1,
    create: 1,
    aggregate: 2,
    findAll: 0,
    findById: 1,
    findOne: 0,
    reload: 0,
    find: 0,
};
exports.BaseModel = BaseModel;
//# sourceMappingURL=BaseModel.js.map