"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const models_1 = require("../services/models");
const association_1 = require("../services/association");
const models_2 = require("../services/models");
const association_2 = require("../services/association");
/**
 * Why does v3/Sequlize and v4/Sequelize does not extend? Because of
 * the transpile target, which is for v3/Sequelize and BaseSequelize ES5
 * and for v4/Sequelize ES6. This is needed for extending the original
 * Sequelize (version 4), which is an ES6 class: ES5 constructor-pattern
 * "classes" cannot extend ES6 classes
 */
class BaseSequelize {
    constructor() {
        this.thoughMap = {};
    }
    static extend(target) {
        // PROTOTYPE MEMBERS
        // --------------------------
        // copies all prototype members of this to target.prototype
        Object
            .getOwnPropertyNames(this.prototype)
            .forEach(name => target.prototype[name] = this.prototype[name]);
        // STATIC MEMBERS
        // --------------------------
        // copies all static members of this to target
        Object
            .keys(this)
            .forEach(name => target[name] = this[name]);
    }
    /**
     * Prepares sequelize config passed to original sequelize constructor
     */
    static prepareConfig(config) {
        if (config.validateOnly) {
            return Object.assign({}, config, {
                name: '_name_',
                username: '_username_',
                password: '_password_',
                dialect: 'sqlite',
                dialectModulePath: __dirname + '/../utils/db-dialect-dummy'
            });
        }
        return config;
    }
    addModels(arg) {
        const models = models_1.getModels(arg);
        this.defineModels(models);
        this.associateModels(models);
        models_2.resolveScopes(models);
    }
    init(config) {
        if (config.modelPaths)
            this.addModels(config.modelPaths);
    }
    /**
     * Processes model associations
     */
    associateModels(models) {
        models.forEach(model => {
            const associations = association_1.getAssociations(model.prototype);
            if (!associations)
                return;
            associations.forEach(association => {
                const foreignKey = association.foreignKey || association_2.getForeignKey(model, association);
                const relatedClass = association.relatedClassGetter();
                let through;
                let otherKey;
                if (association.relation === association_1.BELONGS_TO_MANY) {
                    if (association.otherKey) {
                        otherKey = association.otherKey;
                    }
                    else {
                        if (!association.relatedClassGetter) {
                            throw new Error(`RelatedClassGetter missing on "${model['name']}"`);
                        }
                        otherKey = association_2.getForeignKey(association.relatedClassGetter(), association);
                    }
                    if (association.through) {
                        if (!this.thoughMap[association.through]) {
                            const throughModel = this.getThroughModel(association.through);
                            this.addModels([throughModel]);
                            this.thoughMap[association.through] = throughModel;
                        }
                        through = this.thoughMap[association.through];
                    }
                    else {
                        if (!association.throughClassGetter) {
                            throw new Error(`ThroughClassGetter missing on "${model['name']}"`);
                        }
                        through = association.throughClassGetter();
                    }
                }
                model[association.relation](relatedClass, {
                    as: association.as,
                    through,
                    foreignKey,
                    otherKey
                });
                // The associations has to be adjusted
                const _association = model['associations'][association.as];
                // String based through's need adjustment
                if (association.through) {
                    // as and associationAccessor values referring to string "Through"
                    _association.oneFromSource.as = association.through;
                    _association.oneFromSource.options.as = association.through;
                    _association.oneFromSource.associationAccessor = association.through;
                    _association.oneFromTarget.as = association.through;
                    _association.oneFromTarget.options.as = association.through;
                    _association.oneFromTarget.associationAccessor = association.through;
                }
                this.adjustAssociation(model, _association);
            });
        });
    }
}
exports.BaseSequelize = BaseSequelize;
//# sourceMappingURL=BaseSequelize.js.map